from unittest.mock import Mock, patch

import requests_mock
from bravado.exception import HTTPInternalServerError
from django.core.cache import cache
from django.test import TestCase
from requests.exceptions import HTTPError

from eveuniverse.constants import EveGroupId
from eveuniverse.core import (
    dotlan,
    esitools,
    eveimageserver,
    eveitems,
    evemicros,
    evesdeapi,
    eveskinserver,
    evewho,
    evexml,
    zkillboard,
)
from eveuniverse.models import EveEntity
from eveuniverse.utils import NoSocketsTestCase

from .testdata.esi import EsiClientStub
from .testdata.factories import (
    create_eve_entity,
    create_evemicros_response,
    create_evesdeapi_response,
)


class TestDotlan(TestCase):
    def test_alliance_url(self):
        self.assertEqual(
            dotlan.alliance_url("Wayne Enterprices"),
            "https://evemaps.dotlan.net/alliance/Wayne_Enterprices",
        )

    def test_corporation_url(self):
        self.assertEqual(
            dotlan.corporation_url("Wayne Technology"),
            "https://evemaps.dotlan.net/corp/Wayne_Technology",
        )
        self.assertEqual(
            dotlan.corporation_url("Crédit Agricole"),
            "https://evemaps.dotlan.net/corp/Cr%C3%A9dit_Agricole",
        )

    def test_faction_url(self):
        self.assertEqual(
            dotlan.faction_url("Amarr Empire"),
            "https://evemaps.dotlan.net/factionwarfare/Amarr_Empire",
        )

    def test_region_url(self):
        self.assertEqual(
            dotlan.region_url("Black Rise"), "https://evemaps.dotlan.net/map/Black_Rise"
        )

    def test_solar_system_url(self):
        self.assertEqual(
            dotlan.solar_system_url("Jita"), "https://evemaps.dotlan.net/system/Jita"
        )

    def test_station_url(self):
        self.assertEqual(
            dotlan.station_url("Rakapas V - Home Guard Assembly Plant"),
            "https://evemaps.dotlan.net/station/Rakapas_V_-_Home_Guard_Assembly_Plant",
        )


class TestEveImageServer(TestCase):
    """unit test for eveimageserver"""

    def test_sizes(self):
        self.assertEqual(
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CHARACTER, 42
            ),
            "https://images.evetech.net/characters/42/portrait?size=32",
        )
        self.assertEqual(
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CHARACTER, 42, size=32
            ),
            "https://images.evetech.net/characters/42/portrait?size=32",
        )
        self.assertEqual(
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CHARACTER, 42, size=64
            ),
            "https://images.evetech.net/characters/42/portrait?size=64",
        )
        self.assertEqual(
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CHARACTER, 42, size=128
            ),
            "https://images.evetech.net/characters/42/portrait?size=128",
        )
        self.assertEqual(
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CHARACTER, 42, size=256
            ),
            "https://images.evetech.net/characters/42/portrait?size=256",
        )
        self.assertEqual(
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CHARACTER, 42, size=512
            ),
            "https://images.evetech.net/characters/42/portrait?size=512",
        )
        self.assertEqual(
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CHARACTER, 42, size=1024
            ),
            "https://images.evetech.net/characters/42/portrait?size=1024",
        )
        with self.assertRaises(ValueError):
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CORPORATION, 42, size=-5
            )

        with self.assertRaises(ValueError):
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CORPORATION, 42, size=0
            )

        with self.assertRaises(ValueError):
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CORPORATION, 42, size=31
            )

        with self.assertRaises(ValueError):
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CORPORATION, 42, size=1025
            )

        with self.assertRaises(ValueError):
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CORPORATION, 42, size=2048
            )

    def test_variant(self):
        self.assertEqual(
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CHARACTER,
                42,
                variant=eveimageserver.ImageVariant.PORTRAIT,
            ),
            "https://images.evetech.net/characters/42/portrait?size=32",
        )
        self.assertEqual(
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.ALLIANCE,
                42,
                variant=eveimageserver.ImageVariant.LOGO,
            ),
            "https://images.evetech.net/alliances/42/logo?size=32",
        )
        with self.assertRaises(ValueError):
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CHARACTER,
                42,
                variant=eveimageserver.ImageVariant.LOGO,
            )

    def test_categories(self):
        self.assertEqual(
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.ALLIANCE, 42
            ),
            "https://images.evetech.net/alliances/42/logo?size=32",
        )
        self.assertEqual(
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CORPORATION, 42
            ),
            "https://images.evetech.net/corporations/42/logo?size=32",
        )
        self.assertEqual(
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CHARACTER, 42
            ),
            "https://images.evetech.net/characters/42/portrait?size=32",
        )
        with self.assertRaises(ValueError):
            eveimageserver._eve_entity_image_url("invalid", 42)  # type: ignore

    def test_tenants(self):
        self.assertEqual(
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CHARACTER,
                42,
                tenant=eveimageserver.EsiTenant.TRANQUILITY,
            ),
            "https://images.evetech.net/characters/42/portrait?size=32&tenant=tranquility",
        )
        with self.assertRaises(ValueError):
            eveimageserver._eve_entity_image_url(
                eveimageserver.EsiCategory.CHARACTER, 42, tenant="xxx"  # type: ignore
            )

    def test_alliance_logo_url(self):
        expected = "https://images.evetech.net/alliances/42/logo?size=128"
        self.assertEqual(eveimageserver.alliance_logo_url(42, 128), expected)

    def test_corporation_logo_url(self):
        expected = "https://images.evetech.net/corporations/42/logo?size=128"
        self.assertEqual(eveimageserver.corporation_logo_url(42, 128), expected)

    def test_character_portrait_url(self):
        expected = "https://images.evetech.net/characters/42/portrait?size=128"
        self.assertEqual(eveimageserver.character_portrait_url(42, 128), expected)

    def test_faction_logo_url(self):
        expected = "https://images.evetech.net/corporations/42/logo?size=128"
        self.assertEqual(eveimageserver.faction_logo_url(42, 128), expected)

    def test_type_icon_url(self):
        expected = "https://images.evetech.net/types/42/icon?size=128"
        self.assertEqual(eveimageserver.type_icon_url(42, 128), expected)

    def test_type_render_url(self):
        expected = "https://images.evetech.net/types/42/render?size=128"
        self.assertEqual(eveimageserver.type_render_url(42, 128), expected)

    def test_type_bp_url(self):
        expected = "https://images.evetech.net/types/42/bp?size=128"
        self.assertEqual(eveimageserver.type_bp_url(42, 128), expected)

    def test_type_bpc_url(self):
        expected = "https://images.evetech.net/types/42/bpc?size=128"
        self.assertEqual(eveimageserver.type_bpc_url(42, 128), expected)


class TestEveSkinServer(TestCase):
    """unit test for eveskinserver"""

    def test_default(self):
        """when called without size, will return url with default size"""
        self.assertEqual(
            eveskinserver.type_icon_url(42),
            "https://eveskinserver.kalkoken.net/skin/42/icon?size=32",
        )

    def test_valid_size(self):
        """when called with valid size, will return url with size"""
        self.assertEqual(
            eveskinserver.type_icon_url(42, size=64),
            "https://eveskinserver.kalkoken.net/skin/42/icon?size=64",
        )

    def test_invalid_size(self):
        """when called with invalid size, will raise exception"""
        with self.assertRaises(ValueError):
            eveskinserver.type_icon_url(42, size=22)


@requests_mock.Mocker()
class TestEveMicrosNearestCelestial(TestCase):
    def setUp(self) -> None:
        cache.clear()

    def test_should_return_item_from_api(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET",
            url="https://www.kalkoken.org/apps/evemicros/eveUniverse.php?nearestCelestials=30002682,660502472160,-130687672800,-813545103840",
            json=create_evemicros_response(40170698, 50011472, 40170697),
        )
        # when
        result = evemicros.nearest_celestial(
            solar_system_id=30002682, x=660502472160, y=-130687672800, z=-813545103840
        )
        # then
        self.assertEqual(result.id, 40170698)
        self.assertEqual(result.name, "Colelie VI - Asteroid Belt 1")
        self.assertEqual(result.type_id, 15)
        self.assertEqual(result.distance, 701983769)
        self.assertEqual(requests_mocker.call_count, 1)

    def test_should_return_item_from_cache(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET",
            url="https://www.kalkoken.org/apps/evemicros/eveUniverse.php?nearestCelestials=99,1,2,3",
            json=create_evemicros_response(40170698, 50011472, 40170697),
        )
        evemicros.nearest_celestial(solar_system_id=99, x=1, y=2, z=3)
        # when
        result = evemicros.nearest_celestial(solar_system_id=99, x=1, y=2, z=3)
        # then
        self.assertEqual(result.id, 40170698)
        self.assertEqual(requests_mocker.call_count, 1)

    def test_should_return_none_if_nothing_found(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET",
            url="https://www.kalkoken.org/apps/evemicros/eveUniverse.php?nearestCelestials=30002682,1,2,3",
            json=create_evemicros_response(),
        )
        # when
        result = evemicros.nearest_celestial(solar_system_id=30002682, x=1, y=2, z=3)
        # then
        self.assertIsNone(result)

    def test_should_return_none_if_api_reports_error(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET",
            url="https://www.kalkoken.org/apps/evemicros/eveUniverse.php?nearestCelestials=30002682,1,2,3",
            json=create_evemicros_response(40170698, 50011472, ok=False),
        )
        # when
        result = evemicros.nearest_celestial(solar_system_id=30002682, x=1, y=2, z=3)
        # then
        self.assertIsNone(result)

    def test_should_raise_exception_for_http_errors(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET",
            url="https://www.kalkoken.org/apps/evemicros/eveUniverse.php?nearestCelestials=30002682,1,2,3",
            status_code=500,
        )
        # when
        with self.assertRaises(HTTPError):
            evemicros.nearest_celestial(solar_system_id=30002682, x=1, y=2, z=3)

    def test_should_return_moon_from_api(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET",
            url="https://www.kalkoken.org/apps/evemicros/eveUniverse.php?nearestCelestials=30002682,660502472160,-130687672800,-813545103840",
            json=create_evemicros_response(40170698, 50011472, 40170697, 40170699),
        )
        # when
        result = evemicros.nearest_celestial(
            solar_system_id=30002682,
            x=660502472160,
            y=-130687672800,
            z=-813545103840,
            group_id=EveGroupId.MOON,
        )
        # then
        self.assertEqual(result.id, 40170699)


class TestEveItems(TestCase):
    def test_type_url(self):
        self.assertEqual(
            eveitems.type_url(603), "https://www.kalkoken.org/apps/eveitems/?typeId=603"
        )


class TestEveWho(TestCase):
    def test_alliance_url(self):
        self.assertEqual(
            evewho.alliance_url(12345678), "https://evewho.com/alliance/12345678"
        )

    def test_corporation_url(self):
        self.assertEqual(
            evewho.corporation_url(12345678), "https://evewho.com/corporation/12345678"
        )

    def test_character_url(self):
        self.assertEqual(
            evewho.character_url(12345678), "https://evewho.com/character/12345678"
        )


class TestEveXml(NoSocketsTestCase):
    def test_should_remove_loc_tag_1(self):
        input = "<loc>Character</loc>"
        expected = "Character"
        self.assertHTMLEqual(evexml.remove_loc_tag(input), expected)

    def test_should_remove_loc_tag_2(self):
        input = "Character"
        expected = "Character"
        self.assertHTMLEqual(evexml.remove_loc_tag(input), expected)

    def test_should_detect_url(self):
        self.assertTrue(evexml.is_url("https://www.example.com/bla"))

    def test_should_detect_non_url(self):
        self.assertFalse(evexml.is_url("no-url"))

    @patch("eveuniverse.managers.esi")
    def test_should_convert_links(self, mock_esi):
        # given
        mock_esi.client = EsiClientStub()
        create_eve_entity(
            id=1001, name="Bruce Wayne", category=EveEntity.CATEGORY_CHARACTER
        )
        create_eve_entity(
            id=2001, name="Wayne Technologies", category=EveEntity.CATEGORY_CORPORATION
        )
        create_eve_entity(
            id=3001, name="Wayne Enterprises", category=EveEntity.CATEGORY_ALLIANCE
        )
        create_eve_entity(
            id=30004984, name="Abune", category=EveEntity.CATEGORY_SOLAR_SYSTEM
        )
        create_eve_entity(
            id=60003760,
            name="Jita IV - Moon 4 - Caldari Navy Assembly Plant",
            category=EveEntity.CATEGORY_STATION,
        )
        my_tests = [
            (
                "Alliance",
                "showinfo:16159//3001",
                "https://evemaps.dotlan.net/alliance/Wayne_Enterprises",
            ),
            ("Character", "showinfo:1376//1001", "https://evewho.com/character/1001"),
            (
                "Corporation",
                "showinfo:2//2001",
                "https://evemaps.dotlan.net/corp/Wayne_Technologies",
            ),
            (
                "Killmail",
                "killReport:84900666:9e6fe9e5392ff0cfc6ab956677dbe1deb69c4b04",
                "https://zkillboard.com/kill/84900666/",
            ),
            (
                "Solar System",
                "showinfo:5//30004984",
                "https://evemaps.dotlan.net/system/Abune",
            ),
            (
                "Station",
                "showinfo:52678//60003760",
                "https://evemaps.dotlan.net/station/Jita_IV_-_Moon_4_-_Caldari_Navy_Assembly_Plant",
            ),
            (
                "Inventory Type",
                "showinfo:603",
                "https://www.kalkoken.org/apps/eveitems/?typeId=603",
            ),
            ("Valid URL", "https://www.example.com", "https://www.example.com"),
            (
                "Not support eve link 1",
                "fitting:11987:2048;1:1952;1:26914;2:31366;1:16487;2:31059;1:19057;2:18867;1:18710;1:18871;1:12058;1:31900;1:41155;1::",
                "",
            ),
            (
                "Not support eve link 2",
                "hyperNet:9ff5fa81-942e-49c2-9469-623b2abcb05d",
                "",
            ),
            ("Invalid URL", "not-valid", ""),
            (
                "Unsuported eve links",
                'showinfo:35825//1000000000001">Amamake - Test Structure Alpha',
                "",
            ),
        ]
        with patch("eveuniverse.models.EVEUNIVERSE_LOAD_ASTEROID_BELTS", False), patch(
            "eveuniverse.models.EVEUNIVERSE_LOAD_DOGMAS", False
        ), patch("eveuniverse.models.EVEUNIVERSE_LOAD_GRAPHICS", False), patch(
            "eveuniverse.models.EVEUNIVERSE_LOAD_MARKET_GROUPS", False
        ), patch(
            "eveuniverse.models.EVEUNIVERSE_LOAD_MOONS", False
        ), patch(
            "eveuniverse.models.EVEUNIVERSE_LOAD_PLANETS", False
        ), patch(
            "eveuniverse.models.EVEUNIVERSE_LOAD_STARGATES", False
        ), patch(
            "eveuniverse.models.EVEUNIVERSE_LOAD_STARS", False
        ), patch(
            "eveuniverse.models.EVEUNIVERSE_LOAD_STATIONS", False
        ), patch(
            "eveuniverse.models.EVEUNIVERSE_LOAD_TYPE_MATERIALS", False
        ):
            for test, input, expected in my_tests:
                with self.subTest(test=test):
                    self.assertEqual(evexml.eve_link_to_url(input), expected)


@patch("eveuniverse.core.esitools.esi")
class TestIsEsiOnline(NoSocketsTestCase):
    def test_is_online(self, mock_esi):
        mock_esi.client = EsiClientStub()

        self.assertTrue(esitools.is_esi_online())

    def test_is_offline(self, mock_esi):
        mock_esi.client.Status.get_status.side_effect = HTTPInternalServerError(
            Mock(**{"response.status_code": 500})
        )

        self.assertFalse(esitools.is_esi_online())


class TestZkillboard(TestCase):
    def test_alliance_url(self):
        self.assertEqual(
            zkillboard.alliance_url(12345678),
            "https://zkillboard.com/alliance/12345678/",
        )

    def test_corporation_url(self):
        self.assertEqual(
            zkillboard.corporation_url(12345678),
            "https://zkillboard.com/corporation/12345678/",
        )

    def test_character_url(self):
        self.assertEqual(
            zkillboard.character_url(12345678),
            "https://zkillboard.com/character/12345678/",
        )

    def test_killmail_url(self):
        self.assertEqual(
            zkillboard.killmail_url(12345678), "https://zkillboard.com/kill/12345678/"
        )

    def test_region_url(self):
        self.assertEqual(
            zkillboard.region_url(12345678), "https://zkillboard.com/region/12345678/"
        )

    def test_solar_system_url(self):
        self.assertEqual(
            zkillboard.solar_system_url(12345678),
            "https://zkillboard.com/system/12345678/",
        )


@requests_mock.Mocker()
class TestEveSdeApiNearestCelestial(TestCase):
    _BASE_URL = "https://evesdeapi.kalkoken.net/latest"

    def setUp(self) -> None:
        cache.clear()

    def test_should_return_item_from_api(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET",
            url=(
                f"{self._BASE_URL}/universe/systems/30002682/nearest_celestials"
                "?x=660502472160&y=-130687672800&z=-813545103840"
            ),
            json=create_evesdeapi_response(40170698, 50011472, 40170697),
        )
        # when
        result = evesdeapi.nearest_celestial(
            solar_system_id=30002682, x=660502472160, y=-130687672800, z=-813545103840
        )
        # then
        self.assertEqual(result.id, 40170698)
        self.assertEqual(result.name, "Colelie VI - Asteroid Belt 1")
        self.assertEqual(result.type_id, 15)
        self.assertEqual(result.distance, 701983769)

    def test_should_return_item_from_cache(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET",
            url=(
                f"{self._BASE_URL}/universe/systems/30002682/nearest_celestials"
                "?x=660502472160&y=-130687672800&z=-813545103840"
            ),
            json=create_evesdeapi_response(40170698, 50011472, 40170697),
        )
        evesdeapi.nearest_celestial(
            solar_system_id=30002682, x=660502472160, y=-130687672800, z=-813545103840
        )  # when
        result = evesdeapi.nearest_celestial(
            solar_system_id=30002682, x=660502472160, y=-130687672800, z=-813545103840
        )  # then
        self.assertEqual(result.id, 40170698)
        self.assertEqual(requests_mocker.call_count, 1)

    def test_should_return_none_if_nothing_found(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET",
            url=(
                f"{self._BASE_URL}/universe/systems/30002682/nearest_celestials"
                "?x=1&y=2&z=3"
            ),
            json=create_evesdeapi_response(),
        )
        # when
        result = evesdeapi.nearest_celestial(solar_system_id=30002682, x=1, y=2, z=3)
        # then
        self.assertIsNone(result)

    def test_should_raise_exception_for_http_errors(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET",
            url=(
                f"{self._BASE_URL}/universe/systems/30002682/nearest_celestials"
                "?x=1&y=2&z=3"
            ),
            status_code=500,
        )
        # when
        with self.assertRaises(HTTPError):
            evesdeapi.nearest_celestial(solar_system_id=30002682, x=1, y=2, z=3)

    def test_should_return_moon_from_api(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET",
            url=(
                f"{self._BASE_URL}/universe/systems/30002682/nearest_celestials"
                "?x=660502472160&y=-130687672800&z=-813545103840&group_id=8"
            ),
            json=create_evesdeapi_response(40170699),
        )
        # when
        result = evesdeapi.nearest_celestial(
            solar_system_id=30002682,
            x=660502472160,
            y=-130687672800,
            z=-813545103840,
            group_id=EveGroupId.MOON,
        )
        # then
        self.assertEqual(result.id, 40170699)



from .utils import clean_setting

EVEUNIVERSE_BULK_METHODS_BATCH_SIZE = clean_setting(
    "EVEUNIVERSE_BULK_METHODS_BATCH_SIZE", 500
)
"""Technical parameter defining the maximum number of objects processed per run
of Django batch methods, e.g. bulk_create and bulk_update.
"""

EVEUNIVERSE_API_SDE_URL = clean_setting(
    "EVEUNIVERSE_API_SDE_URL", "https://sde.eve-o.tech/latest"
)
"""URL to a web site providing the SDE tables as JSON files."""

EVEUNIVERSE_LOAD_ASTEROID_BELTS = clean_setting(
    "EVEUNIVERSE_LOAD_ASTEROID_BELTS", False
)
"""When true will automatically load astroid belts with every solar system."""

EVEUNIVERSE_LOAD_DOGMAS = clean_setting("EVEUNIVERSE_LOAD_DOGMAS", False)
"""When true will automatically load dogma, e.g. with every type."""

EVEUNIVERSE_LOAD_GRAPHICS = clean_setting("EVEUNIVERSE_LOAD_GRAPHICS", False)
"""When true will automatically load graphics with every type."""

EVEUNIVERSE_LOAD_MARKET_GROUPS = clean_setting("EVEUNIVERSE_LOAD_MARKET_GROUPS", False)
"""When true will automatically load market groups with every type."""

EVEUNIVERSE_LOAD_MOONS = clean_setting("EVEUNIVERSE_LOAD_MOONS", False)
"""When true will automatically load moons be with every planet."""

EVEUNIVERSE_LOAD_PLANETS = clean_setting("EVEUNIVERSE_LOAD_PLANETS", False)
"""When true will automatically load planets with every solar system."""

EVEUNIVERSE_LOAD_STARGATES = clean_setting("EVEUNIVERSE_LOAD_STARGATES", False)
"""When true will automatically load stargates with every solar system."""

EVEUNIVERSE_LOAD_STARS = clean_setting("EVEUNIVERSE_LOAD_STARS", False)
"""When true will automatically load stars with every solar system."""

EVEUNIVERSE_LOAD_STATIONS = clean_setting("EVEUNIVERSE_LOAD_STATIONS", False)
"""When true will automatically load stations be with every solar system."""

EVEUNIVERSE_LOAD_TYPE_MATERIALS = clean_setting(
    "EVEUNIVERSE_LOAD_TYPE_MATERIALS", False
)
"""When true will automatically load type materials be with every type."""

EVEUNIVERSE_LOAD_INDUSTRY_ACTIVITIES = clean_setting(
    "EVEUNIVERSE_LOAD_INDUSTRY_ACTIVITIES", False
)


EVEUNIVERSE_LOAD_TASKS_PRIORITY = clean_setting("EVEUNIVERSE_LOAD_TASKS_PRIORITY", 6)
"""Priority of tasks for data loads.
This priority should be below 5 to not interfere with normal task operation.
"""

EVEUNIVERSE_REQUESTS_DEFAULT_TIMEOUT = clean_setting(
    "EVEUNIVERSE_REQUESTS_DEFAULT_TIMEOUT", 5
)
"""Default timeout for HTTP GET requests in seconds."""


EVEUNIVERSE_TASKS_TIME_LIMIT = clean_setting("EVEUNIVERSE_TASKS_TIME_LIMIT", 7200)
"""Global timeout for tasks in seconds to reduce task accumulation during outages."""


EVEUNIVERSE_USE_EVESKINSERVER = clean_setting("EVEUNIVERSE_USE_EVESKINSERVER", True)
"""When True a call to EveType.icon_url for a SKIN type will return a eveskinserver URL
else it will return a generic SKIN icon.
"""



from eveuniverse.helpers import (
    EveEntityNameResolver,
    dict_hash,
    get_or_create_esi_or_none,
    meters_to_au,
    meters_to_ly,
)
from eveuniverse.models import EveEntity
from eveuniverse.utils import NoSocketsTestCase

from .testdata.factories import create_eve_entity


class TestHelpers(NoSocketsTestCase):
    def test_meters_to_ly(self):
        self.assertEqual(meters_to_ly(9_460_730_472_580_800), 1)
        self.assertEqual(meters_to_ly(0), 0)
        with self.assertRaises(ValueError):
            meters_to_ly("invalid")

    def test_meters_to_au(self):
        self.assertEqual(meters_to_au(149_597_870_691), 1)
        self.assertEqual(meters_to_au(0), 0)
        with self.assertRaises(ValueError):
            meters_to_au("invalid")


class TestGetOrCreateEsiOrNone(NoSocketsTestCase):
    def test_return_obj_when_property_found(self):
        # given
        create_eve_entity(id=1001, name="Alpha", category="corporation")
        entry = {"type_id": 1001}
        # when
        obj: EveEntity = get_or_create_esi_or_none("type_id", entry, EveEntity)
        # then
        self.assertEqual(obj.id, 1001)

    def test_return_none_when_property_not_found(self):
        # given
        create_eve_entity(id=1001, name="Alpha", category="corporation")
        entry = {}
        # when
        obj = get_or_create_esi_or_none("type_id", entry, EveEntity)
        # then
        self.assertIsNone(obj)


class TestEveEntityNameResolver(NoSocketsTestCase):
    def test_to_name(self):
        resolver = EveEntityNameResolver({1: "alpha", 2: "bravo", 3: "charlie"})
        self.assertEqual(resolver.to_name(2), "bravo")
        self.assertEqual(resolver.to_name(4), "")


class TestDictHash(NoSocketsTestCase):
    def test_should_create_string(self):
        # given
        d1 = {"a": 1, "b": 2, "c": 3}
        d2 = {"a": 1, "b": 2, "c": 4}
        # when
        hash_1a = dict_hash(d1)
        hash_1b = dict_hash(d1)
        hash_2 = dict_hash(d2)
        # then
        self.assertIsInstance(hash_1a, str)
        self.assertEqual(hash_1a, hash_1b)
        self.assertNotEqual(hash_1a, hash_2)



"""Generates profile URLs for dotlan."""

from enum import Enum, auto
from urllib.parse import quote, urljoin


class _Category(Enum):
    ALLIANCE = auto()
    CORPORATION = auto()
    FACTION = auto()
    REGION = auto()
    SOLARSYSTEM = auto()
    STATION = auto()


_BASE_URL = "https://evemaps.dotlan.net"


def _build_url(category: _Category, name: str) -> str:
    """URL to profile page for an eve entity."""

    partials = {
        _Category.ALLIANCE: "alliance",
        _Category.CORPORATION: "corp",
        _Category.FACTION: "factionwarfare",
        _Category.REGION: "map",
        _Category.SOLARSYSTEM: "system",
        _Category.STATION: "station",
    }
    try:
        partial = partials[category]
    except KeyError:
        raise ValueError(f"Invalid category: {category}") from None
    return urljoin(
        _BASE_URL, "{}/{}".format(partial, quote(str(name).replace(" ", "_")))
    )


def alliance_url(name: str) -> str:
    """URL for page about given alliance on dotlan."""
    return _build_url(_Category.ALLIANCE, name)


def corporation_url(name: str) -> str:
    """URL for page about given corporation on dotlan."""
    return _build_url(_Category.CORPORATION, name)


def faction_url(name: str) -> str:
    """URL for page about given corporation on dotlan."""
    return _build_url(_Category.FACTION, name)


def region_url(name: str) -> str:
    """URL for page about given region on dotlan."""
    return _build_url(_Category.REGION, name)


def solar_system_url(name: str) -> str:
    """URL for page about given solar system on dotlan."""
    return _build_url(_Category.SOLARSYSTEM, name)


def station_url(name: str) -> str:
    """URL for page about given solar system on dotlan."""
    return _build_url(_Category.STATION, name)



"""Tasks for Eve Universe."""

import logging
from typing import Iterable, List, Optional

from celery import shared_task
from celery_once import QueueOnce as BaseQueueOnce
from django.db.utils import OperationalError

from . import __title__, models
from .app_settings import EVEUNIVERSE_LOAD_TASKS_PRIORITY, EVEUNIVERSE_TASKS_TIME_LIMIT
from .constants import POST_UNIVERSE_NAMES_MAX_ITEMS, EveCategoryId
from .models import EveEntity, EveMarketPrice, EveType, EveUniverseEntityModel
from .providers import esi
from .utils import LoggerAddTag, chunks

logger = LoggerAddTag(logging.getLogger(__name__), __title__)
# logging.getLogger("esi").setLevel(logging.INFO)


class QueueOnce(BaseQueueOnce):
    """Make sure all redundant tasks will abort gracefully."""

    once = BaseQueueOnce.once
    once["graceful"] = True


# params for all tasks
TASK_DEFAULTS = {"time_limit": EVEUNIVERSE_TASKS_TIME_LIMIT}

# params for tasks that make ESI calls
TASK_ESI_DEFAULTS = {
    **TASK_DEFAULTS,
    **{
        "autoretry_for": [OperationalError],
        "retry_kwargs": {"max_retries": 3},
        "retry_backoff": True,
    },
}
TASK_ESI_DEFAULTS_ONCE = {**TASK_ESI_DEFAULTS, **{"base": QueueOnce}}


# Eve Universe objects


@shared_task(**TASK_ESI_DEFAULTS_ONCE)
def load_eve_object(
    model_name: str, id: int, include_children=False, wait_for_children=True
) -> None:
    """Task for loading an eve object.
    Will only be created from ESI if it does not exist
    """
    logger.info("Loading %s with ID %s", model_name, id)
    ModelClass = EveUniverseEntityModel.get_model_class(model_name)
    ModelClass.objects.get_or_create_esi(  # type: ignore
        id=id, include_children=include_children, wait_for_children=wait_for_children
    )


@shared_task(**TASK_ESI_DEFAULTS_ONCE)
def update_or_create_eve_object(
    model_name: str,
    id: int,
    include_children=False,
    wait_for_children=True,
    enabled_sections: Optional[List[str]] = None,
    task_priority: Optional[int] = None,
) -> None:
    """Update or create an eve object from ESI.

    Args:
        model_name: Name of the respective Django model, e.g. ``"EveType"``
        id: Eve Online ID of object
        include_children: if child objects should be updated/created as well (only when a new object is created)
        wait_for_children: when true child objects will be updated/created blocking (if any), else async (only when a new object is created)
        enabled_sections: Sections to load regardless of current settings, e.g. `[EveType.Section.DOGMAS]` will always load dogmas for EveTypes
        task_priority: priority of started tasks
    """
    logger.info("Updating/Creating %s with ID %s", model_name, id)
    ModelClass = EveUniverseEntityModel.get_model_class(model_name)
    ModelClass.objects.update_or_create_esi(  # type: ignore
        id=id,
        include_children=include_children,
        wait_for_children=wait_for_children,
        enabled_sections=enabled_sections,
        task_priority=task_priority,
    )


@shared_task(**TASK_ESI_DEFAULTS_ONCE)
def update_or_create_inline_object(
    parent_obj_id: int,
    parent_fk: str,
    eve_data_obj: dict,
    other_pk_info: dict,
    parent2_model_name: str,
    inline_model_name: str,
    parent_model_name: str,
    enabled_sections: Optional[List[str]] = None,
) -> None:
    """Task for updating or creating a single inline object from ESI"""
    logger.info(
        "Updating/Creating inline object %s for %s wit ID %s",
        inline_model_name,
        parent_model_name,
        parent_obj_id,
    )
    ModelClass = EveUniverseEntityModel.get_model_class(parent_model_name)
    ModelClass.objects._update_or_create_inline_object(  # type: ignore
        parent_obj_id=parent_obj_id,
        parent_fk=parent_fk,
        eve_data_obj=eve_data_obj,
        other_pk_info=other_pk_info,
        parent2_model_name=parent2_model_name,
        inline_model_name=inline_model_name,
        enabled_sections=enabled_sections,
    )


# EveEntity objects


@shared_task(**TASK_ESI_DEFAULTS)
def create_eve_entities(ids: Iterable[int]) -> None:
    """Task for bulk creating and resolving multiple entities from ESI."""
    EveEntity.objects.bulk_create_esi(ids)


@shared_task(**TASK_ESI_DEFAULTS_ONCE)
def update_unresolved_eve_entities() -> None:
    """Update all unresolved EveEntity objects from ESI.

    Will resolve entities in parallel to speed up resolving large sets.
    """
    ids = list(EveEntity.objects.filter(name="").valid_ids())  # type: ignore
    logger.info("Updating %d unresolved entities from ESI", len(ids))
    for chunk_ids in chunks(ids, POST_UNIVERSE_NAMES_MAX_ITEMS):
        _update_unresolved_eve_entities_for_page.delay(chunk_ids)  # type: ignore


@shared_task(**TASK_ESI_DEFAULTS)
def _update_unresolved_eve_entities_for_page(ids: Iterable[int]) -> None:
    """Update unresolved EveEntity objects for given ids from ESI."""
    EveEntity.objects.update_from_esi_by_id(ids)


# Object loaders


@shared_task(**TASK_ESI_DEFAULTS_ONCE)
def load_map(enabled_sections: Optional[List[str]] = None) -> None:
    """Load the complete Eve map with all regions, constellation and solar systems
    and additional related entities if they are enabled.

    Args:
        enabled_sections: Sections to load regardless of current settings
    """
    logger.info(
        "Loading complete map with all regions, constellations, solar systems "
        "and the following additional entities if related to the map: %s",
        ", ".join(
            EveUniverseEntityModel.determine_effective_sections(enabled_sections)
        ),
    )
    category, method = models.EveRegion._esi_path_list()
    all_ids = getattr(getattr(esi.client, category), method)().results()
    for id in all_ids:
        update_or_create_eve_object.delay(
            model_name="EveRegion",
            id=id,
            include_children=True,
            wait_for_children=False,
            enabled_sections=enabled_sections,
            task_priority=EVEUNIVERSE_LOAD_TASKS_PRIORITY,
        )  # type: ignore


@shared_task(**TASK_ESI_DEFAULTS_ONCE)
def load_all_types(enabled_sections: Optional[List[str]] = None) -> None:
    """Load all eve types.

    Args:
        enabled_sections: Sections to load regardless of current settings
    """
    logger.info(
        "Loading all eve types from ESI including these sections: %s",
        ", ".join(
            EveUniverseEntityModel.determine_effective_sections(enabled_sections)
        ),
    )
    category, method = models.EveCategory._esi_path_list()
    result = getattr(getattr(esi.client, category), method)().results()
    if not result:
        raise ValueError("Did not receive category IDs from ESI.")
    category_ids = sorted(result)
    logger.debug("Fetching categories for IDs: %s", category_ids)
    for category_id in category_ids:
        _load_category_with_children(
            category_id=category_id, enabled_sections=enabled_sections
        )


def _load_category_with_children(
    category_id: int,
    force_loading_dogma: bool = False,
    enabled_sections: Optional[List[str]] = None,
) -> None:
    """Start loading a category async incl. all it's children from ESI."""
    enabled_sections = enabled_sections or []
    if force_loading_dogma:
        enabled_sections.append(EveType.Section.DOGMAS.value)
    update_or_create_eve_object.delay(
        model_name="EveCategory",
        id=category_id,
        include_children=True,
        wait_for_children=False,
        enabled_sections=enabled_sections,
        task_priority=EVEUNIVERSE_LOAD_TASKS_PRIORITY,
    )  # type: ignore


def _load_group_with_children(group_id: int, force_loading_dogma: bool = False) -> None:
    """Starts a task for loading a group incl. all it's children from ESI"""
    enabled_sections = [EveType.Section.DOGMAS.value] if force_loading_dogma else None
    update_or_create_eve_object.delay(
        model_name="EveGroup",
        id=group_id,
        include_children=True,
        wait_for_children=False,
        enabled_sections=enabled_sections,
        task_priority=EVEUNIVERSE_LOAD_TASKS_PRIORITY,
    )  # type: ignore


def _load_type_with_children(type_id: int, force_loading_dogma: bool = False) -> None:
    """Starts a task for loading a type incl. all it's children from ESI"""
    enabled_sections = [EveType.Section.DOGMAS.value] if force_loading_dogma else None
    update_or_create_eve_object.delay(
        model_name="EveType",
        id=type_id,
        include_children=False,
        wait_for_children=False,
        enabled_sections=enabled_sections,
        task_priority=EVEUNIVERSE_LOAD_TASKS_PRIORITY,
    )  # type: ignore


@shared_task(**TASK_ESI_DEFAULTS_ONCE)
def load_ship_types(enabled_sections: Optional[List[str]] = None) -> None:
    """Load all ship types.

    Args:
        enabled_sections: Sections to load regardless of current settings
    """
    logger.info("Started loading all ship types into eveuniverse")
    _load_category_with_children(
        category_id=EveCategoryId.SHIP.value, enabled_sections=enabled_sections
    )


@shared_task(**TASK_ESI_DEFAULTS_ONCE)
def load_structure_types(enabled_sections: Optional[List[str]] = None) -> None:
    """Load all structure types.

    Args:
        enabled_sections: Sections to load regardless of current settings
    """
    logger.info("Started loading all structure types into eveuniverse")
    _load_category_with_children(
        category_id=EveCategoryId.STRUCTURE.value, enabled_sections=enabled_sections
    )


@shared_task(**TASK_ESI_DEFAULTS_ONCE)
def load_eve_types(
    category_ids: Optional[List[int]] = None,
    group_ids: Optional[List[int]] = None,
    type_ids: Optional[List[int]] = None,
    force_loading_dogma: bool = False,
) -> None:
    """Load specified eve types from ESI. Will always load all children except for EveType

    Args:
    - category_ids: EveCategory IDs
    - group_ids: EveGroup IDs
    - type_ids: EveType IDs
    - load_dogma: When True will load dogma for all types
    """
    logger.info("Started loading several eve types into eveuniverse")
    if category_ids:
        for category_id in category_ids:
            _load_category_with_children(category_id, force_loading_dogma)

    if group_ids:
        for group_id in group_ids:
            _load_group_with_children(group_id, force_loading_dogma)

    if type_ids:
        for type_id in type_ids:
            _load_type_with_children(type_id, force_loading_dogma)


@shared_task(**TASK_ESI_DEFAULTS_ONCE)
def update_market_prices(minutes_until_stale: Optional[int] = None):
    """Updates market prices from ESI.
    see EveMarketPrice.objects.update_from_esi() for details"""
    EveMarketPrice.objects.update_from_esi(minutes_until_stale)



EXPECTATION_TEXT = (
    "Note that this process can take some time to complete. "
    "It will create many tasks to load the data and "
    "you can watch the progress on the dashboard. "
    "It is likely finished as soon as the task queue is empty again."
)


def get_input(text):
    """wrapped input to enable unit testing / patching"""
    return input(text)



import logging

from django.core.management.base import BaseCommand

from eveuniverse import __title__, tasks
from eveuniverse.core.esitools import is_esi_online
from eveuniverse.models import EveUniverseEntityModel
from eveuniverse.utils import LoggerAddTag

from . import EXPECTATION_TEXT, get_input

logger = LoggerAddTag(logging.getLogger(__name__), __title__)


class Command(BaseCommand):
    help = (
        "Loads large sets of types as specified from ESI into the local database."
        " This is a helper command meant to be called from other apps only."
    )

    def add_arguments(self, parser):
        parser.add_argument("app_name", help="Name of app this data is loaded for")
        parser.add_argument(
            "--category_id",
            action="append",
            type=int,
            help="Eve category ID to be loaded excl. dogma",
        )
        parser.add_argument(
            "--category_id_with_dogma",
            action="append",
            type=int,
            help="Eve category ID to be loaded incl. dogma",
        )
        parser.add_argument(
            "--group_id",
            action="append",
            type=int,
            help="Eve group ID to be loaded  excl. dogma",
        )
        parser.add_argument(
            "--group_id_with_dogma",
            action="append",
            type=int,
            help="Eve group ID to be loaded incl. dogma",
        )
        parser.add_argument(
            "--type_id",
            action="append",
            type=int,
            help="Eve type ID to be loaded  excl. dogma",
        )
        parser.add_argument(
            "--type_id_with_dogma",
            action="append",
            type=int,
            help="Eve type ID to be loaded  incl. dogma",
        )
        parser.add_argument(
            "--disable_esi_check",
            action="store_true",
            help="Disables checking that ESI is online",
        )
        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_true",
            help="Do NOT prompt the user for input of any kind.",
        )

    def write_to_be_loaded(self, name, *items):
        items_count = sum_items(*items)
        if items_count:
            self.stdout.write(f"{name} to be loaded: {items_count}")

    def handle(self, *args, **options):
        self.stdout.write("Eve Universe - Types Loader")
        self.stdout.write("===========================")

        app_name = options["app_name"]
        category_ids = options["category_id"]
        category_ids_with_dogma = options["category_id_with_dogma"]
        group_ids = options["group_id"]
        group_ids_with_dogma = options["group_id_with_dogma"]
        type_ids = options["type_id"]
        type_ids_with_dogma = options["type_id_with_dogma"]

        if (
            not category_ids
            and not category_ids_with_dogma
            and not group_ids
            and not group_ids_with_dogma
            and not type_ids
            and not type_ids_with_dogma
        ):
            self.stdout.write(self.style.WARNING("No IDs specified. Nothing to do."))
            return

        self.stdout.write("Checking ESI...", ending="")
        if not options["disable_esi_check"] and not is_esi_online():
            self.stdout.write(
                "ESI does not appear to be online at this time. Please try again later."
            )
            self.stdout.write(self.style.WARNING("Aborted"))
            return
        self.stdout.write("ONLINE")
        self.stdout.write(
            f"This command will start loading data for the app: {app_name}."
        )
        self.write_to_be_loaded("Categories", category_ids, category_ids_with_dogma)
        self.write_to_be_loaded("Groups", group_ids, group_ids)
        self.write_to_be_loaded("Types", type_ids, type_ids_with_dogma)
        additional_objects = list(EveUniverseEntityModel.determine_effective_sections())
        if additional_objects:
            self.stdout.write(
                "It will also load the following additional entities when related to "
                "objects loaded for the app: "
                f"{','.join(additional_objects)}"
            )
        self.stdout.write(EXPECTATION_TEXT)
        if not options["noinput"]:
            user_input = get_input("Are you sure you want to proceed? (Y/n)? ")
        else:
            user_input = "y"
        if user_input.lower() != "n":
            if category_ids or group_ids or type_ids:
                tasks.load_eve_types.delay(
                    category_ids=category_ids, group_ids=group_ids, type_ids=type_ids
                )
            if category_ids_with_dogma or group_ids_with_dogma or type_ids_with_dogma:
                tasks.load_eve_types.delay(
                    category_ids=category_ids_with_dogma,
                    group_ids=group_ids_with_dogma,
                    type_ids=type_ids_with_dogma,
                    force_loading_dogma=True,
                )
            self.stdout.write(self.style.SUCCESS("Data load started!"))
        else:
            self.stdout.write(self.style.WARNING("Aborted"))


def sum_items(*items) -> len:
    total = 0
    for item in items:
        try:
            total += len(item)
        except TypeError:
            pass
    return total



# Generated by Django 4.0.3 on 2022-04-11 16:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("eveuniverse", "0006_alter_evedogmaeffectmodifier_operator"),
    ]

    operations = [
        migrations.AddField(
            model_name="evetype",
            name="description",
            field=models.TextField(default=""),
        ),
    ]



import json
from pathlib import Path

import requests

from eveuniverse.app_settings import EVEUNIVERSE_REQUESTS_DEFAULT_TIMEOUT

TYPE_IDS = [
    34,
    35,
    36,
    37,
    38,
    39,
    40,
    3380,
    21947,
    21949,
    21951,
    21953,
    21955,
    21957,
    21959,
    21961,
    21967,
]

print("Fetching ESI raw data...")

# fetching types from ESI
esi_data = {"categories": dict(), "groups": dict(), "types": dict()}
category_ids = set()
group_ids = set()
print("Fetching types...")
for type_id in TYPE_IDS:
    r = requests.get(
        f"https://esi.evetech.net/latest/universe/types/{type_id}/",
        timeout=EVEUNIVERSE_REQUESTS_DEFAULT_TIMEOUT,
    )
    r.raise_for_status()
    info = r.json()
    esi_data["types"][type_id] = info
    group_ids.add(info["group_id"])

print("Fetching groups...")
for group_id in group_ids:
    r = requests.get(
        f"https://esi.evetech.net/latest/universe/groups/{group_id}/",
        timeout=EVEUNIVERSE_REQUESTS_DEFAULT_TIMEOUT,
    )
    r.raise_for_status()
    info = r.json()
    esi_data["groups"][group_id] = info
    category_ids.add(info["category_id"])

print("Fetching categories...")
for category_id in category_ids:
    r = requests.get(
        f"https://esi.evetech.net/latest/universe/categories/{category_id}/",
        timeout=EVEUNIVERSE_REQUESTS_DEFAULT_TIMEOUT,
    )
    r.raise_for_status()
    info = r.json()
    esi_data["categories"][category_id] = info

# writing raw data
esi_data_path = Path(__file__).parent / "esi_raw_data.json"
with esi_data_path.open("w", encoding="utf-8") as fp:
    json.dump(esi_data, fp, indent=4, sort_keys=True)

print("DONE")



import json
from pathlib import Path


def _load_sde_data() -> dict:
    esi_data_path = Path(__file__).parent / "sde_data.json"
    with esi_data_path.open("r", encoding="utf-8") as fp:
        return json.load(fp)


sde_data = _load_sde_data()


def type_materials_cache_content():
    return cache_content(table="type_materials")


def cache_content(table):
    data_all = dict()
    cached_data = sde_data[table]
    for row in cached_data:
        type_id = row["typeID"]
        if type_id not in data_all:
            data_all[type_id] = list()
        data_all[type_id].append(row)
    return data_all



"""Wrapper to access evemicros API."""

from dataclasses import dataclass
from typing import Optional
from urllib.parse import urlencode

import requests
from django.core.cache import cache

from eveuniverse.app_settings import EVEUNIVERSE_REQUESTS_DEFAULT_TIMEOUT

_CACHE_TIMEOUT = 3_600 * 12
_BASE_URL = "https://www.kalkoken.org/apps/evemicros/eveUniverse.php"


@dataclass
class EveItem:
    """A celestial item."""

    id: int
    name: str
    type_id: int
    distance: float


def nearest_celestial(
    solar_system_id: int, x: int, y: int, z: int, group_id: Optional[int] = None
) -> Optional[EveItem]:
    """Fetch nearest celestial to given coordinates from API. Results are cached.

    Args:
        solar_system_id: Eve ID of solar system
        x, y, z: Start point in space to look from
        group_id: Eve ID of group to filter results by

    Raises:
        HTTPError: If an HTTP error is encountered

    Returns:
        Found Eve item or None if nothing found nearby.
    """
    result = _fetch_result_from_api_cached(
        solar_system_id=solar_system_id, x=x, y=y, z=z
    )
    return _get_item_from_result(result, group_id)


def _fetch_result_from_api_cached(
    solar_system_id: int, x: int, y: int, z: int
) -> Optional[dict]:
    """Fetches result from API or cache.
    Returns None if data from API does not have expected structure.
    """
    params = map(str, map(int, [solar_system_id, x, y, z]))
    query = urlencode({"nearestCelestials": ",".join(params)})
    cache_key = f"EVEUNIVERSE_NEAREST_CELESTIAL_{query}"
    result = cache.get(key=cache_key)
    if not result:
        r = requests.get(
            f"{_BASE_URL}?{query}", timeout=EVEUNIVERSE_REQUESTS_DEFAULT_TIMEOUT
        )
        r.raise_for_status()
        data = r.json()
        if "ok" not in data or not data["ok"] or "result" not in data:
            return None
        result = data["result"]
        cache.set(key=cache_key, value=result, timeout=_CACHE_TIMEOUT)
    return result


def _get_item_from_result(result, group_id) -> Optional[EveItem]:
    """Tries to find item in result. Returns None if item can not be found."""
    if not result:
        return None
    if not group_id:
        return _create_item(result[0])
    for item in result:
        if item["groupID"] == group_id:
            return _create_item(item)
    return None


def _create_item(record: dict) -> EveItem:
    return EveItem(
        id=int(record["itemID"]),
        name=str(record["itemName"]),
        type_id=int(record["typeID"]),
        distance=float(record["distanceKm"]),
    )



import unittest
from unittest.mock import patch

from django.test.utils import override_settings

from eveuniverse.constants import EveCategoryId
from eveuniverse.models import (
    EsiMapping,
    EveAncestry,
    EveBloodline,
    EveCategory,
    EveConstellation,
    EveDogmaAttribute,
    EveDogmaEffect,
    EveEntity,
    EveGraphic,
    EveGroup,
    EveMarketGroup,
    EveRegion,
    EveType,
    EveTypeDogmaEffect,
    EveUnit,
    EveUniverseEntityModel,
)
from eveuniverse.utils import NoSocketsTestCase

from .testdata.esi import EsiClientStub

unittest.util._MAX_LENGTH = 1000
MODELS_PATH = "eveuniverse.models"
MANAGERS_PATH = "eveuniverse.managers"


@patch(MANAGERS_PATH + ".esi")
class TestEveType(NoSocketsTestCase):
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_GRAPHICS", False)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", False)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", False)
    def test_can_create_type_from_esi_excluding_all(self, mock_esi):
        mock_esi.client = EsiClientStub()

        obj, created = EveType.objects.get_or_create_esi(id=603)
        self.assertTrue(created)
        self.assertEqual(obj.id, 603)
        self.assertEqual(obj.name, "Merlin")
        self.assertEqual(
            obj.description,
            """The Merlin is the most powerful combat frigate of the Caldari. Its role has evolved through the years, and while its defenses have always remained exceptionally strong for a Caldari vessel, its offensive capabilities have evolved from versatile, jack-of-all-trades attack patterns into focused and deadly gunfire tactics. The Merlin's primary aim is to have its turrets punch holes in opponents' hulls.""",
        )
        self.assertEqual(obj.capacity, 150)
        self.assertEqual(obj.eve_group, EveGroup.objects.get(id=25))
        self.assertEqual(obj.mass, 997000)
        self.assertEqual(obj.packaged_volume, 2500)
        self.assertEqual(obj.portion_size, 1)
        self.assertTrue(obj.published)
        self.assertEqual(obj.radius, 39)
        self.assertEqual(obj.volume, 16500)
        self.assertIsNone(obj.eve_graphic)
        self.assertIsNone(obj.eve_market_group)
        self.assertEqual(obj.dogma_attributes.count(), 0)
        self.assertEqual(obj.dogma_effects.count(), 0)
        self.assertEqual(obj.eve_entity_category(), EveEntity.CATEGORY_INVENTORY_TYPE)

    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_GRAPHICS", True)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", True)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", True)
    def test_can_create_type_from_esi_including_dogmas(self, mock_esi):
        mock_esi.client = EsiClientStub()

        eve_type, created = EveType.objects.get_or_create_esi(id=603)
        self.assertTrue(created)
        self.assertEqual(eve_type.id, 603)
        self.assertEqual(eve_type.eve_graphic, EveGraphic.objects.get(id=314))
        self.assertEqual(eve_type.eve_market_group, EveMarketGroup.objects.get(id=61))

        dogma_attribute_1 = eve_type.dogma_attributes.filter(
            eve_dogma_attribute=EveDogmaAttribute.objects.get(id=588)
        ).first()
        self.assertEqual(dogma_attribute_1.value, 5)
        dogma_attribute_1 = eve_type.dogma_attributes.filter(
            eve_dogma_attribute=EveDogmaAttribute.objects.get(id=129)
        ).first()
        self.assertEqual(dogma_attribute_1.value, 12)

        dogma_effect_1 = eve_type.dogma_effects.filter(
            eve_dogma_effect=EveDogmaEffect.objects.get(id=1816)
        ).first()
        self.assertFalse(dogma_effect_1.is_default)
        dogma_effect_2 = eve_type.dogma_effects.filter(
            eve_dogma_effect=EveDogmaEffect.objects.get(id=1817)
        ).first()
        self.assertTrue(dogma_effect_2.is_default)

    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", True)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", False)
    def test_when_disabled_can_create_type_from_esi_excluding_dogmas(self, mock_esi):
        mock_esi.client = EsiClientStub()

        obj, created = EveType.objects.get_or_create_esi(id=603)
        self.assertTrue(created)
        self.assertEqual(obj.id, 603)
        self.assertTrue(obj.eve_market_group, EveMarketGroup.objects.get(id=61))
        self.assertEqual(obj.dogma_attributes.count(), 0)
        self.assertEqual(obj.dogma_effects.count(), 0)

    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", False)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", True)
    def test_when_disabled_can_create_type_from_esi_excluding_market_groups(
        self, mock_esi
    ):
        mock_esi.client = EsiClientStub()

        eve_type, created = EveType.objects.get_or_create_esi(id=603)
        self.assertTrue(created)
        self.assertEqual(eve_type.id, 603)
        self.assertIsNone(eve_type.eve_market_group)
        self.assertSetEqual(
            set(
                eve_type.dogma_attributes.values_list(
                    "eve_dogma_attribute_id", flat=True
                )
            ),
            {588, 129},
        )
        self.assertSetEqual(
            set(eve_type.dogma_effects.values_list("eve_dogma_effect_id", flat=True)),
            {1816, 1817},
        )

    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_GRAPHICS", False)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", False)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", False)
    def test_can_create_type_from_esi_including_dogmas_when_disabled_1(self, mock_esi):
        mock_esi.client = EsiClientStub()

        eve_type, created = EveType.objects.update_or_create_esi(
            id=603, enabled_sections=[EveType.LOAD_DOGMAS]
        )
        self.assertTrue(created)
        self.assertEqual(eve_type.id, 603)
        self.assertSetEqual(
            set(
                eve_type.dogma_attributes.values_list(
                    "eve_dogma_attribute_id", flat=True
                )
            ),
            {588, 129},
        )
        self.assertSetEqual(
            set(eve_type.dogma_effects.values_list("eve_dogma_effect_id", flat=True)),
            {1816, 1817},
        )

    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_GRAPHICS", False)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", False)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", False)
    def test_can_create_type_from_esi_including_dogmas_when_disabled_2(self, mock_esi):
        mock_esi.client = EsiClientStub()

        eve_type, created = EveType.objects.get_or_create_esi(
            id=603, enabled_sections=[EveType.LOAD_DOGMAS]
        )
        self.assertTrue(created)
        self.assertEqual(eve_type.id, 603)
        self.assertSetEqual(
            set(
                eve_type.dogma_attributes.values_list(
                    "eve_dogma_attribute_id", flat=True
                )
            ),
            {588, 129},
        )
        self.assertSetEqual(
            set(eve_type.dogma_effects.values_list("eve_dogma_effect_id", flat=True)),
            {1816, 1817},
        )

    @override_settings(
        CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True
    )
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_GRAPHICS", False)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", False)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", False)
    def test_can_create_type_from_esi_including_children_as_task(self, mock_esi):
        # given
        mock_esi.client = EsiClientStub()
        # when
        eve_type, created = EveType.objects.update_or_create_esi(
            id=603, wait_for_children=False, enabled_sections=[EveType.LOAD_DOGMAS]
        )
        # then
        self.assertTrue(created)
        self.assertEqual(eve_type.id, 603)
        self.assertSetEqual(
            set(
                eve_type.dogma_attributes.values_list(
                    "eve_dogma_attribute_id", flat=True
                )
            ),
            {588, 129},
        )
        self.assertSetEqual(
            set(eve_type.dogma_effects.values_list("eve_dogma_effect_id", flat=True)),
            {1816, 1817},
        )

    @override_settings(
        CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True
    )
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_GRAPHICS", False)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", False)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", False)
    def test_can_create_type_from_esi_including_children_as_task_with_priority(
        self, mock_esi
    ):
        # given
        mock_esi.client = EsiClientStub()
        # when
        eve_type, created = EveType.objects.update_or_create_esi(
            id=603,
            wait_for_children=False,
            enabled_sections=[EveType.LOAD_DOGMAS],
            task_priority=7,
        )
        # then
        self.assertTrue(created)
        self.assertEqual(eve_type.id, 603)
        self.assertSetEqual(
            set(
                eve_type.dogma_attributes.values_list(
                    "eve_dogma_attribute_id", flat=True
                )
            ),
            {588, 129},
        )
        self.assertSetEqual(
            set(eve_type.dogma_effects.values_list("eve_dogma_effect_id", flat=True)),
            {1816, 1817},
        )

    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", False)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", False)
    def test_can_create_render_url(self, mock_esi):
        mock_esi.client = EsiClientStub()

        eve_type, created = EveType.objects.get_or_create_esi(id=603)
        self.assertTrue(created)
        self.assertEqual(
            eve_type.render_url(256),
            "https://images.evetech.net/types/603/render?size=256",
        )


@patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", False)
@patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", False)
@patch(MANAGERS_PATH + ".esi")
class TestEveTypeIconUrl(NoSocketsTestCase):
    def test_can_create_icon_url_1(self, mock_esi):
        """icon from regular type, automatically detected"""
        mock_esi.client = EsiClientStub()
        eve_type, _ = EveType.objects.get_or_create_esi(id=603)

        self.assertEqual(
            eve_type.icon_url(256), "https://images.evetech.net/types/603/icon?size=256"
        )

    def test_can_create_icon_url_2(self, mock_esi):
        """icon from blueprint type, automatically detected"""
        mock_esi.client = EsiClientStub()
        eve_type, _ = EveType.objects.get_or_create_esi(id=950)

        self.assertEqual(
            eve_type.icon_url(256), "https://images.evetech.net/types/950/bp?size=256"
        )

    def test_can_create_icon_url_3(self, mock_esi):
        """icon from regular type, preset as blueprint"""
        mock_esi.client = EsiClientStub()
        eve_type, _ = EveType.objects.get_or_create_esi(id=603)

        self.assertEqual(
            eve_type.icon_url(size=256, is_blueprint=True),
            "https://images.evetech.net/types/603/bp?size=256",
        )

    def test_can_create_icon_url_3a(self, mock_esi):
        """icon from regular type, preset as blueprint"""
        mock_esi.client = EsiClientStub()
        eve_type, _ = EveType.objects.get_or_create_esi(id=603)

        self.assertEqual(
            eve_type.icon_url(size=256, category_id=EveCategoryId.BLUEPRINT),
            "https://images.evetech.net/types/603/bp?size=256",
        )

    @patch(MODELS_PATH + ".EVEUNIVERSE_USE_EVESKINSERVER", False)
    def test_can_create_icon_url_5(self, mock_esi):
        """when called for SKIN type, will return dummy SKIN URL with requested size"""
        mock_esi.client = EsiClientStub()
        eve_type, _ = EveType.objects.get_or_create_esi(id=34599)

        self.assertIn("skin_generic_64.png", eve_type.icon_url(size=64))

    @patch(MODELS_PATH + ".EVEUNIVERSE_USE_EVESKINSERVER", False)
    def test_can_create_icon_url_5a(self, mock_esi):
        """when called for SKIN type, will return dummy SKIN URL with requested size"""
        mock_esi.client = EsiClientStub()
        eve_type, _ = EveType.objects.get_or_create_esi(id=34599)

        self.assertIn("skin_generic_32.png", eve_type.icon_url(size=32))

    @patch(MODELS_PATH + ".EVEUNIVERSE_USE_EVESKINSERVER", False)
    def test_can_create_icon_url_5b(self, mock_esi):
        """when called for SKIN type, will return dummy SKIN URL with requested size"""
        mock_esi.client = EsiClientStub()
        eve_type, _ = EveType.objects.get_or_create_esi(id=34599)

        self.assertIn("skin_generic_128.png", eve_type.icon_url(size=128))

    @patch(MODELS_PATH + ".EVEUNIVERSE_USE_EVESKINSERVER", False)
    def test_can_create_icon_url_5c(self, mock_esi):
        """when called for SKIN type and size is invalid, then raise exception"""
        mock_esi.client = EsiClientStub()
        eve_type, _ = EveType.objects.get_or_create_esi(id=34599)

        with self.assertRaises(ValueError):
            eve_type.icon_url(size=512)

        with self.assertRaises(ValueError):
            eve_type.icon_url(size=1024)

        with self.assertRaises(ValueError):
            eve_type.icon_url(size=31)

    @patch(MODELS_PATH + ".EVEUNIVERSE_USE_EVESKINSERVER", False)
    def test_can_create_icon_url_6(self, mock_esi):
        """when called for non SKIN type and SKIN is forced, then return SKIN URL"""
        mock_esi.client = EsiClientStub()
        eve_type, _ = EveType.objects.get_or_create_esi(id=950)

        self.assertIn(
            "skin_generic_128.png",
            eve_type.icon_url(size=128, category_id=EveCategoryId.SKIN),
        )

    @patch(MODELS_PATH + ".EVEUNIVERSE_USE_EVESKINSERVER", False)
    def test_can_create_icon_url_7(self, mock_esi):
        """when called for SKIN type and regular is forced, then return regular URL"""
        mock_esi.client = EsiClientStub()
        eve_type, _ = EveType.objects.get_or_create_esi(id=34599)

        self.assertEqual(
            eve_type.icon_url(size=256, category_id=EveCategoryId.STRUCTURE),
            "https://images.evetech.net/types/34599/icon?size=256",
        )

    @patch(MODELS_PATH + ".EVEUNIVERSE_USE_EVESKINSERVER", True)
    def test_can_create_icon_url_8(self, mock_esi):
        """
        when called for SKIN type and eveskinserver is enabled,
        then return corresponding eveskinserver URL
        """
        mock_esi.client = EsiClientStub()
        eve_type, _ = EveType.objects.get_or_create_esi(id=34599)

        self.assertEqual(
            eve_type.icon_url(size=256),
            "https://eveskinserver.kalkoken.net/skin/34599/icon?size=256",
        )

    @patch(MODELS_PATH + ".EVEUNIVERSE_USE_EVESKINSERVER", True)
    def test_can_create_icon_url_9(self, mock_esi):
        """can use variants"""
        mock_esi.client = EsiClientStub()
        eve_type, _ = EveType.objects.get_or_create_esi(id=603)

        self.assertEqual(
            eve_type.icon_url(size=256, variant=EveType.IconVariant.REGULAR),
            "https://images.evetech.net/types/603/icon?size=256",
        )
        self.assertEqual(
            eve_type.icon_url(size=256, variant=EveType.IconVariant.BPO),
            "https://images.evetech.net/types/603/bp?size=256",
        )
        self.assertEqual(
            eve_type.icon_url(size=256, variant=EveType.IconVariant.BPC),
            "https://images.evetech.net/types/603/bpc?size=256",
        )
        self.assertEqual(
            eve_type.icon_url(size=256, variant=EveType.IconVariant.SKIN),
            "https://eveskinserver.kalkoken.net/skin/603/icon?size=256",
        )


@patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", False)
@patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", False)
@patch(MANAGERS_PATH + ".esi")
class TestEveTypeProfileUrl(NoSocketsTestCase):
    def test_can_url(self, mock_esi):
        # given
        mock_esi.client = EsiClientStub()
        eve_type, _ = EveType.objects.get_or_create_esi(id=603)
        # when
        result = eve_type.profile_url
        # then
        self.assertEqual(result, "https://www.kalkoken.org/apps/eveitems/?typeId=603")


class TestEveUnit(NoSocketsTestCase):
    def test_get_object(self):
        obj = EveUnit.objects.get(id=10)
        self.assertEqual(obj.id, 10)
        self.assertEqual(obj.name, "Speed")


class TestEsiMapping(NoSocketsTestCase):
    maxDiff = None

    def test_single_pk(self):
        mapping = EveCategory._esi_mapping()
        self.assertEqual(len(mapping.keys()), 3)
        self.assertEqual(
            mapping["id"],
            EsiMapping(
                esi_name="category_id",
                is_optional=False,
                is_pk=True,
                is_fk=False,
                related_model=None,
                is_parent_fk=False,
                is_charfield=False,
                create_related=True,
            ),
        )
        self.assertEqual(
            mapping["name"],
            EsiMapping(
                esi_name="name",
                is_optional=True,
                is_pk=False,
                is_fk=False,
                related_model=None,
                is_parent_fk=False,
                is_charfield=True,
                create_related=True,
            ),
        )
        self.assertEqual(
            mapping["published"],
            EsiMapping(
                esi_name="published",
                is_optional=False,
                is_pk=False,
                is_fk=False,
                related_model=None,
                is_parent_fk=False,
                is_charfield=False,
                create_related=True,
            ),
        )

    def test_with_fk(self):
        mapping = EveConstellation._esi_mapping()
        self.assertEqual(len(mapping.keys()), 6)
        self.assertEqual(
            mapping["id"],
            EsiMapping(
                esi_name="constellation_id",
                is_optional=False,
                is_pk=True,
                is_fk=False,
                related_model=None,
                is_parent_fk=False,
                is_charfield=False,
                create_related=True,
            ),
        )
        self.assertEqual(
            mapping["name"],
            EsiMapping(
                esi_name="name",
                is_optional=True,
                is_pk=False,
                is_fk=False,
                related_model=None,
                is_parent_fk=False,
                is_charfield=True,
                create_related=True,
            ),
        )
        self.assertEqual(
            mapping["eve_region"],
            EsiMapping(
                esi_name="region_id",
                is_optional=False,
                is_pk=False,
                is_fk=True,
                related_model=EveRegion,
                is_parent_fk=False,
                is_charfield=False,
                create_related=True,
            ),
        )
        self.assertEqual(
            mapping["position_x"],
            EsiMapping(
                esi_name=("position", "x"),
                is_optional=True,
                is_pk=False,
                is_fk=False,
                related_model=None,
                is_parent_fk=False,
                is_charfield=False,
                create_related=True,
            ),
        )
        self.assertEqual(
            mapping["position_y"],
            EsiMapping(
                esi_name=("position", "y"),
                is_optional=True,
                is_pk=False,
                is_fk=False,
                related_model=None,
                is_parent_fk=False,
                is_charfield=False,
                create_related=True,
            ),
        )
        self.assertEqual(
            mapping["position_z"],
            EsiMapping(
                esi_name=("position", "z"),
                is_optional=True,
                is_pk=False,
                is_fk=False,
                related_model=None,
                is_parent_fk=False,
                is_charfield=False,
                create_related=True,
            ),
        )

    def test_optional_fields(self):
        mapping = EveAncestry._esi_mapping()
        self.assertEqual(len(mapping.keys()), 6)
        self.assertEqual(
            mapping["id"],
            EsiMapping(
                esi_name="id",
                is_optional=False,
                is_pk=True,
                is_fk=False,
                related_model=None,
                is_parent_fk=False,
                is_charfield=False,
                create_related=True,
            ),
        )
        self.assertEqual(
            mapping["name"],
            EsiMapping(
                esi_name="name",
                is_optional=True,
                is_pk=False,
                is_fk=False,
                related_model=None,
                is_parent_fk=False,
                is_charfield=True,
                create_related=True,
            ),
        )
        self.assertEqual(
            mapping["eve_bloodline"],
            EsiMapping(
                esi_name="bloodline_id",
                is_optional=False,
                is_pk=False,
                is_fk=True,
                related_model=EveBloodline,
                is_parent_fk=False,
                is_charfield=False,
                create_related=True,
            ),
        )
        self.assertEqual(
            mapping["description"],
            EsiMapping(
                esi_name="description",
                is_optional=False,
                is_pk=False,
                is_fk=False,
                related_model=None,
                is_parent_fk=False,
                is_charfield=True,
                create_related=True,
            ),
        )
        self.assertEqual(
            mapping["icon_id"],
            EsiMapping(
                esi_name="icon_id",
                is_optional=True,
                is_pk=False,
                is_fk=False,
                related_model=None,
                is_parent_fk=False,
                is_charfield=False,
                create_related=True,
            ),
        )
        self.assertEqual(
            mapping["short_description"],
            EsiMapping(
                esi_name="short_description",
                is_optional=True,
                is_pk=False,
                is_fk=False,
                related_model=None,
                is_parent_fk=False,
                is_charfield=True,
                create_related=True,
            ),
        )

    def test_inline_model(self):
        mapping = EveTypeDogmaEffect._esi_mapping()
        self.assertEqual(len(mapping.keys()), 3)
        self.assertEqual(
            mapping["eve_type"],
            EsiMapping(
                esi_name="eve_type",
                is_optional=False,
                is_pk=True,
                is_fk=True,
                related_model=EveType,
                is_parent_fk=True,
                is_charfield=False,
                create_related=True,
            ),
        )
        self.assertEqual(
            mapping["eve_dogma_effect"],
            EsiMapping(
                esi_name="effect_id",
                is_optional=False,
                is_pk=True,
                is_fk=True,
                related_model=EveDogmaEffect,
                is_parent_fk=False,
                is_charfield=False,
                create_related=True,
            ),
        )
        self.assertEqual(
            mapping["is_default"],
            EsiMapping(
                esi_name="is_default",
                is_optional=False,
                is_pk=False,
                is_fk=False,
                related_model=None,
                is_parent_fk=False,
                is_charfield=False,
                create_related=True,
            ),
        )

    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_GRAPHICS", True)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", True)
    @patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", True)
    def test_EveType_mapping(self):
        mapping = EveType._esi_mapping()
        self.assertSetEqual(
            set(mapping.keys()),
            {
                "id",
                "name",
                "description",
                "capacity",
                "eve_group",
                "eve_graphic",
                "icon_id",
                "eve_market_group",
                "mass",
                "packaged_volume",
                "portion_size",
                "radius",
                "published",
                "volume",
            },
        )


class TestDetermineEnabledSections(NoSocketsTestCase):
    def test_should_return_empty_1(self):
        # when
        with patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_ASTEROID_BELTS", False), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", False
        ), patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_GRAPHICS", False), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_MOONS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_PLANETS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_STARGATES", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_STARS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_STATIONS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_TYPE_MATERIALS", False
        ):
            result = EveUniverseEntityModel.determine_effective_sections()
        # then
        self.assertSetEqual(result, set())

    def test_should_return_empty_2(self):
        # when
        with patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_ASTEROID_BELTS", False), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", False
        ), patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_GRAPHICS", False), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_MOONS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_PLANETS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_STARGATES", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_STARS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_STATIONS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_TYPE_MATERIALS", False
        ):
            result = EveUniverseEntityModel.determine_effective_sections(None)
        # then
        self.assertSetEqual(result, set())

    def test_should_return_global_section(self):
        # when
        with patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_ASTEROID_BELTS", False), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", True
        ), patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_GRAPHICS", False), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_MOONS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_PLANETS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_STARGATES", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_STARS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_STATIONS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_TYPE_MATERIALS", False
        ):
            result = EveUniverseEntityModel.determine_effective_sections()
        # then
        self.assertSetEqual(result, {EveType.Section.DOGMAS})

    def test_should_combine_global_and_local_sections(self):
        # when
        with patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_ASTEROID_BELTS", False), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_DOGMAS", True
        ), patch(MODELS_PATH + ".EVEUNIVERSE_LOAD_GRAPHICS", False), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_MARKET_GROUPS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_MOONS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_PLANETS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_STARGATES", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_STARS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_STATIONS", False
        ), patch(
            MODELS_PATH + ".EVEUNIVERSE_LOAD_TYPE_MATERIALS", False
        ):
            result = EveUniverseEntityModel.determine_effective_sections(
                ["type_materials"]
            )
        # then
        self.assertSetEqual(
            result, {EveType.Section.DOGMAS, EveType.Section.TYPE_MATERIALS}
        )



"""Wrapper for accessing eveskinserver API."""

_EVE_SKIN_SERVER_URL = "https://eveskinserver.kalkoken.net"
_DEFAULT_IMAGE_SIZE = 32


def type_icon_url(type_id: int, size: int = _DEFAULT_IMAGE_SIZE) -> str:
    """icon image URL for the given SKIN type ID"""
    if not size or size < 32 or size > 1024 or (size & (size - 1) != 0):
        raise ValueError(f"Invalid size: {size}")
    return f"{_EVE_SKIN_SERVER_URL}/skin/{int(type_id)}/icon?size={size}"



"""Helper functions and classes."""

import hashlib
import json
from typing import Any, Dict, Optional

from django.db import models


def meters_to_ly(value: float) -> Optional[float]:
    """Convert meters into lightyears."""
    return float(value) / 9_460_730_472_580_800 if value is not None else None


def meters_to_au(value: float) -> Optional[float]:
    """Convert meters into AU."""
    return float(value) / 149_597_870_691 if value is not None else None


def get_or_create_esi_or_none(
    prop_name: str, dct: dict, Model: type
) -> Optional[models.Model]:
    """Create a new eveuniverse object from a dictionary entry and return it
    or return None if the prop name is not in the dict.

    :meta private:
    """
    if eve_id := dct.get(prop_name):
        return Model.objects.get_or_create_esi(id=eve_id)[0]  # type: ignore
    return None


class EveEntityNameResolver:
    """Container with a mapping between entity Ids and entity names
    and a performant API
    """

    def __init__(self, names_map: Dict[int, str]) -> None:
        self._names_map = names_map

    def to_name(self, id: int) -> str:
        """Resolved an entity ID to a name

        Args:
            id: ID of the Eve entity to resolve

        Returns:
            name for corresponding entity ID if known else an empty string
        """
        try:
            name = self._names_map[id]
        except KeyError:
            name = ""

        return name


def dict_hash(dictionary: Dict[str, Any]) -> str:
    """SHA256 hash of a dictionary.

    :meta private:
    """
    my_hash = hashlib.sha256()
    encoded = json.dumps(dictionary, sort_keys=True).encode(encoding="utf8")
    my_hash.update(encoded)
    return my_hash.hexdigest()



import json
from pathlib import Path

import requests

print("Generating SDE data...")

# fetch type materials from SDE bridge
r = requests.get("https://sde.eve-o.tech/latest/invTypeMaterials.json")
r.raise_for_status()
data_all = r.json()

# fetch type IDs from current test data
esi_data_path = Path(__file__).parent / "esi_data.json"
with esi_data_path.open("r", encoding="utf-8") as fp:
    esi_data = json.load(fp)

# extract needed items only
existing_type_ids = set(
    map(int, esi_data["Universe"]["get_universe_types_type_id"].keys())
)
data_excerpt = [row for row in data_all if row["typeID"] in existing_type_ids]
sde_data = {"type_materials": data_excerpt}

# write sde data
sde_data_path = Path(__file__).parent / "sde_data.json"
with sde_data_path.open("w", encoding="utf-8") as fp:
    json.dump(sde_data, fp, indent=4, sort_keys=True)

# check data consistency
missing_type_ids = []
material_type_ids = {row["materialTypeID"] for row in data_excerpt}
for type_id in material_type_ids:
    if type_id not in existing_type_ids:
        missing_type_ids.append(type_id)

print(f"ERROR: Missing type IDs {', '.join(map(str, sorted(material_type_ids)))}")
print("DONE")
